import torch
import torch.nn as nn
class SkipGramModel(torch.nn.Module):
    # Constructor
    def __init__(self, totalNodes, embedDim):
        super(SkipGramModel, self).__init__()
        self.layers(totalNodes, embedDim)

    # Layers for a skipgram model
    def layers(self, totalNodes, embedDim):
        self.W1  = nn.Parameter(torch.rand((totalNodes, embedDim), requires_grad=True))
        self.W2 = nn.Parameter(torch.rand((embedDim, totalNodes), requires_grad=True))

    # Training skipgram model
    def forward(self, features):
        hidden = torch.matmul(features, self.W1)
        embed = torch.matmul(hidden, self.W2)
        # Get embedding
        return embed


import torch
import numpy as np
from collections import defaultdict
import warnings
import sys
try:
    from utils import operator_hadamard, custom_formatwarning
    from randomWalkEmbedding import RandomWalkEmbedding
except ModuleNotFoundError:
    from .utils import operator_hadamard, custom_formatwarning
    from .randomWalkEmbedding import RandomWalkEmbedding

class Node2vec(RandomWalkEmbedding):
    # Constructor
    def __init__(self, graph = None, walkLength = 0, embedDim = 0, numbOfWalksPerVertex = 0, \
                 windowSize = 0, lr = 0, p = 0, q = 0):
        if graph is None:
            warnings.warn("Provide a graph: {}".format(graph))
            sys.exit()
        super(Node2vec, self).__init__(graph, walkLength, embedDim, numbOfWalksPerVertex,  windowSize, lr)

        # validate arguments
        if p == 0:
            self.p = 0.5
            warnings.warn("Set p to default: {}".format(self.p))
        else:
            self.p = p
        if q == 0:
            self.q = 0.8
            warnings.warn("Set q to default: {}".format(self.q))
        else:
            self.q = q
        self.model = None

    # Calculating Probabilities for traversing a node
    def computeProbabilities(self, source_node):
        probs = defaultdict(dict)
        probs[source_node]['probabilities'] = dict()
        for current_node in self.graph.neighbors(source_node):
            probs_ = list()
            for destination in self.graph.neighbors(current_node):
                if source_node == destination:
                    prob_ = self.graph[current_node][destination].get('weight',1) * (1/self.p)
                elif destination in self.graph.neighbors(source_node):
                    prob_ = self.graph[current_node][destination].get('weight',1)
                else:
                    prob_ = self.graph[current_node][destination].get('weight',1) * (1/self.q)
                probs_.append(prob_)
            probs[source_node]['probabilities'][current_node] = probs_/np.sum(probs_)
        return probs

    # Walks generation
    def RandomWalk(self, startNode, walkLength):
        # walk contains encoded node labels
        walk = [int(self.nodeEncoder.transform([startNode]))]
        walkOptions = list(self.graph[startNode])
        if len(walkOptions)==0:
            return walk
        firstStep = np.random.choice(walkOptions)
        walk.append(int(self.nodeEncoder.transform([firstStep])))
        for k in range(walkLength-2):
            walkOptions = list(self.graph[(self.nodeEncoder.inverse_transform([walk[-1]])[0])])
            if len(walkOptions)==0:
                break
            probs = self.computeProbabilities((self.nodeEncoder.inverse_transform([walk[-2]])[0]))
            probabilities = probs[(self.nodeEncoder.inverse_transform([walk[-2]])[0])] \
                ['probabilities'][(self.nodeEncoder.inverse_transform([walk[-1]])[0])]
            # Traverse a vertex from a neighbors of node 'node"
            nextStep = np.random.choice(walkOptions, p=probabilities)
            walk.append(int(self.nodeEncoder.transform([nextStep])))
        return walk

    # Generate features for nodes
    def generateNodeFeatures(self, totalNodes, wvi, j):
        nodeFeatures = torch.zeros(totalNodes)
        nodeFeatures[wvi[j]] = 1
        return nodeFeatures

    # Training graph embedding model
    def learnEmbedding(self, model, wvi):
        for j in range(len(wvi)):
            for k in range(max(0,j-self.windowSize) , min(j+self.windowSize, len(wvi))):
                # generate features
                nodeFeatures = self.generateNodeFeatures(self.totalNodes, wvi, j)
                out = self.model.forward(nodeFeatures)
                loss = torch.log(torch.sum(torch.exp(out))) - out[wvi[k]]
                loss.backward()
                for param in self.model.parameters():
                    param.data.sub_(self.lr*param.grad)
                    param.grad.data.zero_()
        return self.model

    # Get node embedding for a specific node, i.e., "node"
    def getNodeEmbedding(self, node):
        return self.model.W1[int(self.nodeEncoder.transform([node]))].data

    # Training node embedding model
    def learnNodeEmbedding(self, model):
        self.model = model
        for startNode in list(self.graph.nodes):
            for i in range(self.numbOfWalksPerVertex):
                walkStartNode = self.RandomWalk(startNode, self.walkLength)
                self.model = self.learnEmbedding(self.model, walkStartNode)
        return self.model

    # Training edge embedding model
    def learnEdgeEmbedding(self, model):
        self.model = model
        for startNode in list(self.graph.nodes):
            for i in range(self.numbOfWalksPerVertex):
                walkStartNode = self.RandomWalk(startNode, self.walkLength)
                self.model = self.learnEmbedding(self.model, walkStartNode)
        return self.model

    # Get edge embedding for a specific edge having source node, i.e., "srcNode" and destination node, i.e., dstNode
    def getEdgeEmbedding(self, srcNode, dstNode):
        # Operator_hadamrd defined in Utils
        return operator_hadamard(self.getNodeEmbedding(srcNode), self.getNodeEmbedding(dstNode))






import os
import shutil
from collections import ChainMap, deque
import numpy as np
import pandas as pd
from gensim.models import Word2Vec
from joblib import Parallel, delayed
import torch
import warnings
import sys
try:
    from utils import partition_dict, preprocess_nxgraph, get_vertices, create_alias_table, cost, cost_max, \
        compute_dtw_dist, convert_dtw_struc_dist, operator_hadamard, custom_formatwarning
    from biasedRandomWalk import BiasedWalker
    from randomWalkEmbedding import RandomWalkEmbedding
except ModuleNotFoundError:
    from .utils import partition_dict, preprocess_nxgraph, get_vertices, create_alias_table, cost, cost_max, \
        compute_dtw_dist, convert_dtw_struc_dist, operator_hadamard, custom_formatwarning
    from .biasedRandomWalk import BiasedWalker
    from .randomWalkEmbedding import RandomWalkEmbedding

class Struc2Vec(RandomWalkEmbedding):
    # Constructor
    def __init__(self, graph = None, walkLength = 0, embedDim = 0, numbOfWalksPerVertex = 0, \
                 windowSize = 0, lr= 0, verbose=0, stay_prob=0, opt1_reduce_len=True, opt2_reduce_sim_calc=True, \
                 opt3_num_layers=None, temp_path='./temp_struc2vec/', reuse=False):
        if graph is None:
            warnings.warn("Provide a graph: {}".format(graph))
            sys.exit()

        super(Struc2Vec, self).__init__(graph, walkLength, embedDim, numbOfWalksPerVertex,  windowSize, lr)
        # validate arguments
        if stay_prob == 0:
            self.stay_prob = 0.3
            warnings.warn("Set stay prob. to default: {}".format(self.stay_prob))
        else:
            self.stay_prob = stay_prob
        self.verbose = verbose
        self.idx2node, self.node2idx = preprocess_nxgraph(graph)

        self.adj_list, self.nodeEncoder = self.graph_to_adjList(graph)
        self.idx = list(range(len(self.nodeEncoder.classes_)))

        self.opt1_reduce_len = opt1_reduce_len
        self.opt2_reduce_sim_calc = opt2_reduce_sim_calc
        self.opt3_num_layers = opt3_num_layers

        self.resue = reuse
        self.temp_path = temp_path

        if not os.path.exists(self.temp_path):
            os.mkdir(self.temp_path)
        if not reuse:
            shutil.rmtree(self.temp_path)
            os.mkdir(self.temp_path)

        workers = 1
        self.create_context_graph(self.opt3_num_layers, workers, verbose)
        self.prepare_biased_walk()
        self.walker = BiasedWalker(self.nodeEncoder.classes_, self.temp_path)
        self._embeddings = {}

    # Walks generation
    def RandomWalk(self, startNode, walkLength):
        startNode = int(self.nodeEncoder.transform([startNode]))
        walk = self.walker. simulate_walks(
            startNode, walkLength, self.stay_prob, 1, self.verbose)
        return list(self.nodeEncoder.transform(walk[0]))

    # Generate features for nodes
    def generateNodeFeatures(self, totalNodes, wvi, j):
        nodeFeatures = torch.zeros(totalNodes)
        nodeFeatures[wvi[j]] = 1
        return nodeFeatures

    # Training graph embedding model
    def learnEmbedding(self, model, wvi):
        for j in range(len(wvi)):
            for k in range(max(0,j-self.windowSize) , min(j+self.windowSize, len(wvi))):
                # generate features
                nodeFeatures = self.generateNodeFeatures(self.totalNodes, wvi, j)
                out = self.model.forward(nodeFeatures)
                loss = torch.log(torch.sum(torch.exp(out))) - out[wvi[k]]
                loss.backward()
                for param in self.model.parameters():
                    param.data.sub_(self.lr*param.grad)
                    param.grad.data.zero_()
        return self.model

    # Training node embedding model
    def learnNodeEmbedding(self, model):
        self.model = model
        for startNode in list(self.graph.nodes):
            for i in range(self.numbOfWalksPerVertex):
                walkStartNode = self.RandomWalk(startNode, self.walkLength)
                self.model = self.learnEmbedding(self.model, walkStartNode)
        return self.model

    # Get node embedding for a specific node, i.e., "node"
    def getNodeEmbedding(self, node):
        return self.model.W1[int(self.nodeEncoder.transform([node]))].data

    # Training graph embedding model
    def learnEdgeEmbedding(self, model):
        self.model = model
        for startNode in list(self.graph.nodes):
            for i in range(self.numbOfWalksPerVertex):
                walkStartNode = self.RandomWalk(startNode, self.walkLength)
                self.model = self.learnEmbedding(self.model, walkStartNode)
        return self.model

    # Get edge embedding for a specific edge having source node, i.e., "srcNode" and destination node, i.e., dstNode
    def getEdgeEmbedding(self, srcNode, dstNode):
        return operator_hadamard(self.getNodeEmbedding(srcNode), self.getNodeEmbedding(dstNode))

    # Create a context graph
    def create_context_graph(self, max_num_layers, workers=1, verbose=0,):

        pair_distances = self._compute_structural_distance(
            max_num_layers, workers, verbose,)
        layers_adj, layers_distances = self._get_layer_rep(pair_distances)
        pd.to_pickle(layers_adj, self.temp_path + 'layers_adj.pkl')

        layers_accept, layers_alias = self._get_transition_probs(
            layers_adj, layers_distances)
        pd.to_pickle(layers_alias, self.temp_path + 'layers_alias.pkl')
        pd.to_pickle(layers_accept, self.temp_path + 'layers_accept.pkl')

    def prepare_biased_walk(self,):

        sum_weights = {}
        sum_edges = {}
        average_weight = {}
        gamma = {}
        layer = 0
        while (os.path.exists(self.temp_path+'norm_weights_distance-layer-' + str(layer)+'.pkl')):
            probs = pd.read_pickle(
                self.temp_path+'norm_weights_distance-layer-' + str(layer)+'.pkl')
            for v, list_weights in probs.items():
                sum_weights.setdefault(layer, 0)
                sum_edges.setdefault(layer, 0)
                sum_weights[layer] += sum(list_weights)
                sum_edges[layer] += len(list_weights)

            average_weight[layer] = sum_weights[layer] / sum_edges[layer]

            gamma.setdefault(layer, {})

            for v, list_weights in probs.items():
                num_neighbours = 0
                for w in list_weights:
                    if (w > average_weight[layer]):
                        num_neighbours += 1
                gamma[layer][v] = num_neighbours

            layer += 1

        pd.to_pickle(average_weight, self.temp_path + 'average_weight')
        pd.to_pickle(gamma, self.temp_path + 'gamma.pkl')

    def train(self, window_size=5, workers=3, epochs=5):

        # pd.read_pickle(self.temp_path+'walks.pkl')
        sentences = self.sentences

        print("Learning representation...")
        model = Word2Vec(sentences, vector_size=self.embedDim, window=window_size, min_count=0, hs=1, sg=1, workers=workers,
                         epochs=epochs)
        print("Learning representation done!")
        self.w2v_model = model

        return model

    def get_embeddings(self,):
        if self.w2v_model is None:
            print("model not train")
            return {}

        self._embeddings = {}
        for word in self.graph.nodes():
            self._embeddings[word] = self.w2v_model.wv[word]

        return self._embeddings

    def _compute_ordered_degreelist(self, max_num_layers):

        degreeList = {}
        vertices = self.idx  # self.g.nodes()
        for v in vertices:
            degreeList[v] = self._get_order_degreelist_node(v, max_num_layers)
        return degreeList

    def _get_order_degreelist_node(self, root, max_num_layers=None):
        if max_num_layers is None:
            max_num_layers = float('inf')

        ordered_degree_sequence_dict = {}
        visited = [False] * len(self.graph.nodes())
        queue = deque()
        level = 0
        queue.append(root)
        visited[root] = True

        while (len(queue) > 0 and level <= max_num_layers):

            count = len(queue)
            if self.opt1_reduce_len:
                degree_list = {}
            else:
                degree_list = []
            while (count > 0):

                top = queue.popleft()
                node = int(self.nodeEncoder.inverse_transform([top]))
                degree = len(self.graph[node])

                if self.opt1_reduce_len:
                    degree_list[degree] = degree_list.get(degree, 0) + 1
                else:
                    degree_list.append(degree)

                for nei in self.graph[node]:

                    nei_idx = int(self.nodeEncoder.transform([nei]))
                    if not visited[nei_idx]:
                        visited[nei_idx] = True
                        queue.append(nei_idx)
                count -= 1
            if self.opt1_reduce_len:
                orderd_degree_list = [(degree, freq)
                                      for degree, freq in degree_list.items()]
                orderd_degree_list.sort(key=lambda x: x[0])
            else:
                orderd_degree_list = sorted(degree_list)
            ordered_degree_sequence_dict[level] = orderd_degree_list
            level += 1

        return ordered_degree_sequence_dict

    def _compute_structural_distance(self, max_num_layers, workers=1, verbose=0,):

        if os.path.exists(self.temp_path+'structural_dist.pkl'):
            structural_dist = pd.read_pickle(
                self.temp_path+'structural_dist.pkl')
        else:
            if self.opt1_reduce_len:
                dist_func = cost_max
            else:
                dist_func = cost

            if os.path.exists(self.temp_path + 'degreelist.pkl'):
                degreeList = pd.read_pickle(self.temp_path + 'degreelist.pkl')
            else:
                degreeList = self._compute_ordered_degreelist(max_num_layers)
                pd.to_pickle(degreeList, self.temp_path + 'degreelist.pkl')

            if self.opt2_reduce_sim_calc:
                degrees = self._create_vectors()
                degreeListsSelected = {}
                vertices = {}
                n_nodes = len(self.idx)
                for v in self.idx:  # c:list of vertex
                    nbs = get_vertices(
                        v, len(self.graph[int(self.nodeEncoder.inverse_transform([v]))]), degrees, n_nodes)
                    vertices[v] = nbs  # store nbs
                    degreeListsSelected[v] = degreeList[v]  # store dist
                    for n in nbs:
                        # store dist of nbs
                        degreeListsSelected[n] = degreeList[n]
            else:
                vertices = {}
                for v in degreeList:
                    vertices[v] = [vd for vd in degreeList.keys() if vd > v]

            results = Parallel(n_jobs=workers, verbose=verbose,)(
                delayed(compute_dtw_dist)(part_list, degreeList, dist_func) for part_list in partition_dict(vertices, workers))
            dtw_dist = dict(ChainMap(*results))

            structural_dist = convert_dtw_struc_dist(dtw_dist)
            pd.to_pickle(structural_dist, self.temp_path +
                         'structural_dist.pkl')

        return structural_dist

    def _create_vectors(self):
        degrees = {}  # sotre v list of degree
        degrees_sorted = set()  # store degree
        G = self.graph
        for v in self.idx:
            degree = len(G[ int(self.nodeEncoder.inverse_transform([v])) ])
            degrees_sorted.add(degree)
            if (degree not in degrees):
                degrees[degree] = {}
                degrees[degree]['vertices'] = []
            degrees[degree]['vertices'].append(v)
        degrees_sorted = np.array(list(degrees_sorted), dtype='int')
        degrees_sorted = np.sort(degrees_sorted)

        l = len(degrees_sorted)
        for index, degree in enumerate(degrees_sorted):
            if (index > 0):
                degrees[degree]['before'] = degrees_sorted[index - 1]
            if (index < (l - 1)):
                degrees[degree]['after'] = degrees_sorted[index + 1]

        return degrees

    def _get_layer_rep(self, pair_distances):
        layer_distances = {}
        layer_adj = {}
        for v_pair, layer_dist in pair_distances.items():
            for layer, distance in layer_dist.items():
                vx = v_pair[0]
                vy = v_pair[1]

                layer_distances.setdefault(layer, {})
                layer_distances[layer][vx, vy] = distance

                layer_adj.setdefault(layer, {})
                layer_adj[layer].setdefault(vx, [])
                layer_adj[layer].setdefault(vy, [])
                layer_adj[layer][vx].append(vy)
                layer_adj[layer][vy].append(vx)

        return layer_adj, layer_distances

    def _get_transition_probs(self, layers_adj, layers_distances):
        layers_alias = {}
        layers_accept = {}

        for layer in layers_adj:

            neighbors = layers_adj[layer]
            layer_distances = layers_distances[layer]
            node_alias_dict = {}
            node_accept_dict = {}
            norm_weights = {}

            for v, neighbors in neighbors.items():
                e_list = []
                sum_w = 0.0

                for n in neighbors:
                    if (v, n) in layer_distances:
                        wd = layer_distances[v, n]
                    else:
                        wd = layer_distances[n, v]
                    w = np.exp(-float(wd))
                    e_list.append(w)
                    sum_w += w

                e_list = [x / sum_w for x in e_list]
                norm_weights[v] = e_list
                accept, alias = create_alias_table(e_list)
                node_alias_dict[v] = alias
                node_accept_dict[v] = accept

            pd.to_pickle(
                norm_weights, self.temp_path + 'norm_weights_distance-layer-' + str(layer)+'.pkl')

            layers_alias[layer] = node_alias_dict
            layers_accept[layer] = node_accept_dict

        return layers_accept, layers_alias



from skipgram import SkipGramModel
from utils import plot_2DEmbedding
from node2vec import Node2vec
from struc2vec import Struc2Vec
from deepWalk import DeepWalk
from utils import saveEmbedding, loadGraph

# Set Path to data
dataset = "cora"
dataset = "cora - Copy"
data_dir = "..\cora"


# input graph
my_graph = loadGraph(data_dir, dataset)

# Set Parameters (For fast completion we kept the parameters simple)
embedDim = 2 # embedding size
numbOfWalksPerVertex = 2 # walks per vertex
walkLength = 4 # walk lenght
lr =0.25 # learning rate
windowSize = 3 # window size


# DeepWalk
dw = DeepWalk(my_graph, walkLength=walkLength, embedDim=embedDim, numbOfWalksPerVertex=numbOfWalksPerVertex, \
              windowSize=windowSize, lr = lr)

# # Node2Vec
# dw = Node2vec(my_graph, walkLength=walkLength, embedDim=embedDim, numbOfWalksPerVertex=numbOfWalksPerVertex, \
#               windowSize=windowSize, lr=lr, p = 0.5, q = 0.8)
# #
# # Struc2Vec
# dw = Struc2Vec(my_graph, walkLength=walkLength, embedDim=embedDim, numbOfWalksPerVertex=numbOfWalksPerVertex, \
#                windowSize=windowSize, lr = lr, stay_prob=0.3)


# Skip Gram model
model_skip_gram = SkipGramModel(dw.totalNodes, dw.embedDim)

print("Learning Embedding")
# Learning Node Embedding
model = dw.learnNodeEmbedding(model_skip_gram)

# Learning Edge Embedding
# model = dw.learnEdgeEmbedding(model_skip_gram)

# Plot Embedding
plot_2DEmbedding(dw)

# Save embedding to disk
print("Saving embedding to disk")
saveEmbedding(data_dir, dataset, dw)

print("Generating embedding for a node and edge")
node1 = 35
node2 = 40
# Get Embedding for a node
emb = dw.getNodeEmbedding(node1)
print("Node Embedding", emb.data)
#
# # Get Embedding for an edge
emb = dw.getEdgeEmbedding(node1, node2)
print("Edge Embedding", emb.data)






import torch
import random
import warnings
import sys

try:
    from utils import operator_hadamard, custom_formatwarning
    from randomWalkEmbedding import RandomWalkEmbedding
except ModuleNotFoundError:
    from .utils import operator_hadamard, custom_formatwarning
    from .randomWalkEmbedding import RandomWalkEmbedding

class DeepWalk(RandomWalkEmbedding):
    # constructor
    def __init__(self, graph = None, walkLength = 0, embedDim = 0, numbOfWalksPerVertex = 0, \
                 windowSize = 0, lr = 0):
        if graph is None:
            warnings.warn("Provide a graph: {}".format(graph))
            sys.exit()
        super(DeepWalk, self).__init__(graph, walkLength, embedDim, numbOfWalksPerVertex,  windowSize, lr)
        self.model = None

    # Walks generation
    def RandomWalk(self, node,t):
        # walk contains encoded node labels
        walk = [int(self.nodeEncoder.transform([node]))]        # Walk starts from node "node"
        for i in range(t-1):
            neighborsList = [n for n in self.graph.neighbors(node)]
            # Randomly traverse a vertex from a neighbors of node 'node"
            node = neighborsList[random.randint(0,len(neighborsList)-1)]
            walk.append(int(self.nodeEncoder.transform([node])))
        return walk

    # Generate features for nodes
    def generateNodeFeatures(self, totalNodes, wvi, j):
        nodeFeatures = torch.zeros(totalNodes)
        nodeFeatures[wvi[j]] = 1
        return nodeFeatures

    # Training graph embedding model
    def learnEmbedding(self, wvi):
        for j in range(len(wvi)):
            for k in range(max(0,j-self.windowSize) , min(j+self.windowSize, len(wvi))):
                # Getting node for a specific features
                nodeFeatures = self.generateNodeFeatures(self.totalNodes, wvi, j)
                out = self.model.forward(nodeFeatures)
                loss = torch.log(torch.sum(torch.exp(out))) - out[wvi[k]]
                loss.backward()
                for param in self.model.parameters():
                    param.data.sub_(self.lr*param.grad)
                    param.grad.data.zero_()
        return self.model

    # Training node embedding model
    def learnNodeEmbedding(self, model):
        self.model = model
        nodesList = list(self.graph.nodes)
        # Number of walks for a single vertex
        for i in range(self.numbOfWalksPerVertex):
            random.shuffle(nodesList)
            # Generating walk for a vertex
            for vi in nodesList:
                walkStartNode = self.RandomWalk(vi, self.walkLength)
                self.model = self.learnEmbedding(walkStartNode)
        return self.model

    # Get node embedding for a specific node, i.e., "node"
    def getNodeEmbedding(self, node):
        return self.model.W1[int(self.nodeEncoder.transform([node]))].data

    # Training edge embedding model
    def learnEdgeEmbedding(self, model):
        self.model = model
        for startNode in list(self.graph.nodes):
            # Number of walks for a single vertex
            for i in range(self.numbOfWalksPerVertex):
                # Generating walk for a vertex
                walkStartNode = self.RandomWalk(startNode, self.walkLength)
                self.model = self.learnEmbedding(walkStartNode)
        return self.model

    # Get edge embedding for a specific edge having source node, i.e., "srcNode" and destination node, i.e., dstNode
    def getEdgeEmbedding(self, srcNode, dstNode):
        # Operator_hadamrd defined in Utils
        return operator_hadamard(self.getNodeEmbedding(srcNode), self.getNodeEmbedding(dstNode))



import pandas as pd
import numpy as np
import math
from fastdtw import fastdtw
import os
import networkx as nx
import matplotlib.pyplot as plt
import warnings
def chooseNeighbor(v, graphs, layers_alias, layers_accept, layer):

    v_list = graphs[layer][v]

    idx = alias_sample(layers_accept[layer][v], layers_alias[layer][v])
    v = v_list[idx]

    return v

def alias_sample(accept, alias):
    """
    :param accept:
    :param alias:
    :return: sample index
    """
    N = len(accept)
    i = int(np.random.random()*N)
    r = np.random.random()
    if r < accept[i]:
        return i
    else:
        return alias[i]

def preprocess_nxgraph(graph):
    node2idx = {}
    idx2node = []
    node_size = 0
    for node in graph.nodes():
        node2idx[node] = node_size
        idx2node.append(node)
        node_size += 1
    return idx2node, node2idx

def partition_dict(vertices, workers):
    batch_size = (len(vertices) - 1) // workers + 1
    part_list = []
    part = []
    count = 0
    for v1, nbs in vertices.items():
        part.append((v1, nbs))
        count += 1
        if count % batch_size == 0:
            part_list.append(part)
            part = []
    if len(part) > 0:
        part_list.append(part)
    return part_list

def create_alias_table(area_ratio):
    l = len(area_ratio)
    accept, alias = [0] * l, [0] * l
    small, large = [], []
    area_ratio_ = np.array(area_ratio) * l
    for i, prob in enumerate(area_ratio_):
        if prob < 1.0:
            small.append(i)
        else:
            large.append(i)

    while small and large:
        small_idx, large_idx = small.pop(), large.pop()
        accept[small_idx] = area_ratio_[small_idx]
        alias[small_idx] = large_idx
        area_ratio_[large_idx] = area_ratio_[large_idx] - \
                                 (1 - area_ratio_[small_idx])
        if area_ratio_[large_idx] < 1.0:
            small.append(large_idx)
        else:
            large.append(large_idx)

    while large:
        large_idx = large.pop()
        accept[large_idx] = 1
    while small:
        small_idx = small.pop()
        accept[small_idx] = 1

    return accept, alias


def cost(a, b):
    ep = 0.5
    m = max(a, b) + ep
    mi = min(a, b) + ep
    return ((m / mi) - 1)


def cost_min(a, b):
    ep = 0.5
    m = max(a[0], b[0]) + ep
    mi = min(a[0], b[0]) + ep
    return ((m / mi) - 1) * min(a[1], b[1])


def cost_max(a, b):
    ep = 0.5
    m = max(a[0], b[0]) + ep
    mi = min(a[0], b[0]) + ep
    return ((m / mi) - 1) * max(a[1], b[1])


def convert_dtw_struc_dist(distances, startLayer=1):
    for vertices, layers in distances.items():
        keys_layers = sorted(layers.keys())
        startLayer = min(len(keys_layers), startLayer)
        for layer in range(0, startLayer):
            keys_layers.pop(0)

        for layer in keys_layers:
            layers[layer] += layers[layer - 1]
    return distances


def get_vertices(v, degree_v, degrees, n_nodes):
    a_vertices_selected = 2 * math.log(n_nodes, 2)
    vertices = []
    try:
        c_v = 0

        for v2 in degrees[degree_v]['vertices']:
            if (v != v2):
                vertices.append(v2)  # same degree
                c_v += 1
                if (c_v > a_vertices_selected):
                    raise StopIteration

        if ('before' not in degrees[degree_v]):
            degree_b = -1
        else:
            degree_b = degrees[degree_v]['before']
        if ('after' not in degrees[degree_v]):
            degree_a = -1
        else:
            degree_a = degrees[degree_v]['after']
        if (degree_b == -1 and degree_a == -1):
            raise StopIteration  # not anymore v
        degree_now = verifyDegrees(degrees, degree_v, degree_a, degree_b)
        # nearest valid degree
        while True:
            for v2 in degrees[degree_now]['vertices']:
                if (v != v2):
                    vertices.append(v2)
                    c_v += 1
                    if (c_v > a_vertices_selected):
                        raise StopIteration

            if (degree_now == degree_b):
                if ('before' not in degrees[degree_b]):
                    degree_b = -1
                else:
                    degree_b = degrees[degree_b]['before']
            else:
                if ('after' not in degrees[degree_a]):
                    degree_a = -1
                else:
                    degree_a = degrees[degree_a]['after']

            if (degree_b == -1 and degree_a == -1):
                raise StopIteration

            degree_now = verifyDegrees(degrees, degree_v, degree_a, degree_b)

    except StopIteration:
        return list(vertices)

    return list(vertices)


def verifyDegrees(degrees, degree_v_root, degree_a, degree_b):

    if(degree_b == -1):
        degree_now = degree_a
    elif(degree_a == -1):
        degree_now = degree_b
    elif(abs(degree_b - degree_v_root) < abs(degree_a - degree_v_root)):
        degree_now = degree_b
    else:
        degree_now = degree_a

    return degree_now

def compute_dtw_dist(part_list, degreeList, dist_func):
    dtw_dist = {}
    for v1, nbs in part_list:
        lists_v1 = degreeList[v1]  # lists_v1 :orderd degree list of v1
        for v2 in nbs:
            lists_v2 = degreeList[v2]  # lists_v1 :orderd degree list of v2
            max_layer = min(len(lists_v1), len(lists_v2))  # valid layer
            dtw_dist[v1, v2] = {}
            for layer in range(0, max_layer):
                dist, path = fastdtw(
                    lists_v1[layer], lists_v2[layer], radius=1, dist=dist_func)
                dtw_dist[v1, v2][layer] = dist
    return dtw_dist

def operator_hadamard(u, v):
    return u * v

def plot_2DEmbedding(dw):
    xs = dw.model.W1.data[:, 0]
    ys = dw.model.W1.data[:, 1]
    ls = list(range(0, len(xs)))
    plt.scatter(xs, ys)
    for x,y,l in zip(xs,ys, ls):
        plt.annotate((dw.nodeEncoder.inverse_transform([l])[0]), (x, y))
    plt.show()

# Save embedding to disk
def saveEmbedding(data_dir, dataset, dw):
    embCol = list()
    for i in range(dw.embedDim):
        embCol.append('c' + str(i))
    df = pd.DataFrame(columns = embCol,
                      index = list(dw.graph.nodes())) # adding nodes as an index
    for node in dw.graph.nodes():
        # Get embedding for a node
        f = dw.getNodeEmbedding(node)
        df.loc[node] = f
        df.to_csv("{}{}.embedding".format(data_dir + "/", dataset), sep='\t', header=False)

def loadGraph(data_dir, dataset):
    # Load Data
    print("Loading Data...")
    data_dir = os.path.expanduser(data_dir)
    edgelist = pd.read_csv(os.path.join(data_dir, dataset + ".cites"), sep='\t', header=None, names=["target", "source"])
    return nx.from_pandas_edgelist(edgelist)

# custom format warning with only a message
def custom_formatwarning(msg, *args, **kwargs):
    return str(msg) + '\n'
warnings.formatwarning = custom_formatwarning




import pandas as pd
import random
import math

try:
    from utils import chooseNeighbor
except ModuleNotFoundError:
    from .utils import chooseNeighbor

class BiasedWalker:
    def __init__(self, idx2node, temp_path):

        self.idx2node = idx2node
        self.idx = list(range(len(self.idx2node)))
        self.temp_path = temp_path
        pass

    def simulate_walks(self, startNode, walkLength, stay_prob=0.3, workers=1, verbose=0):

        layers_adj = pd.read_pickle(self.temp_path+'layers_adj.pkl')
        layers_alias = pd.read_pickle(self.temp_path+'layers_alias.pkl')
        layers_accept = pd.read_pickle(self.temp_path+'layers_accept.pkl')
        gamma = pd.read_pickle(self.temp_path+'gamma.pkl')
        walks =  self._simulate_walks(startNode, walkLength, stay_prob, layers_adj, layers_accept, layers_alias, gamma)
        return walks

    def _simulate_walks(self, startNode, walkLength, stay_prob, layers_adj, layers_accept, layers_alias, gamma):
        walks = []
        walks.append(self._exec_random_walk(layers_adj, layers_accept, layers_alias,
                                            startNode, walkLength, gamma, stay_prob))
        return walks

    def _exec_random_walk(self, graphs, layers_accept, layers_alias, v, walkLength, gamma, stay_prob=0.3):
        initialLayer = 0
        layer = initialLayer

        path = []
        path.append(self.idx2node[v])

        while len(path) < walkLength:
            r = random.random()
            if(r < stay_prob):  # same layer
                v = chooseNeighbor(v, graphs, layers_alias,
                                   layers_accept, layer)
                path.append(self.idx2node[v])
            else:  # different layer
                r = random.random()
                try:
                    x = math.log(gamma[layer][v] + math.e)
                    p_moveup = (x / (x + 1))
                except:
                    print(layer, v)
                    raise ValueError()

                if(r > p_moveup):
                    if(layer > initialLayer):
                        layer = layer - 1
                else:
                    if((layer + 1) in graphs and v in graphs[layer + 1]):
                        layer = layer + 1

        return path


from abc import ABC, abstractmethod
from sklearn import preprocessing
import warnings
try:
    from utils import custom_formatwarning
except ModuleNotFoundError:
    from .utils import custom_formatwarning

class RandomWalkEmbedding:
    def __init__(self, graph, walkLength, embedDim, numbOfWalksPerVertex, windowSize, lr):
        self.graph = graph
        # validate arguments
        if walkLength == 0:
            self.walkLength = 3
            warnings.warn("Set Walk to default: {}".format(self.walkLength))
        else:
            self.walkLength = walkLength

        if embedDim == 0:
            self.embedDim = 2
            warnings.warn("Set Embedding Dimention to default: {}".format(self.embedDim))
        else:
            self.embedDim = embedDim

        if numbOfWalksPerVertex == 3:
            self.numbOfWalksPerVertex = 3
            warnings.warn("Setting Learning Rate to default: {}".format(self.numbOfWalksPerVertex))
        else:
            self.numbOfWalksPerVertex = numbOfWalksPerVertex

        if windowSize == 0:
            self.windowSize = 3
            warnings.warn("Set Context Window to default: {}".format(self.windowSize))
        else:
            self.windowSize = windowSize

        if lr == 0:
            self.lr = 0.25
            warnings.warn("Set Learning Rate to default: {}".format(self.lr))
        else:
            self.lr = lr
        self.adj_list, self.nodeEncoder = self.graph_to_adjList(graph)
        self.totalNodes = graph.number_of_nodes()




    def encoder(self, graph):
        nodeEncoder = preprocessing.LabelEncoder()
        return nodeEncoder.fit(list(graph.nodes()))

    def graph_to_adjList(self, graph):
        nodeEncoder = self.encoder(graph)
        adj_list1 = [None] * graph.number_of_nodes()
        for node, edges in list(graph.adjacency()):

            #     print(node, list(edges.keys()))
            adj_list1[nodeEncoder.transform([node])[0]] = list(nodeEncoder.transform(list(edges.keys())))
        return adj_list1, nodeEncoder

    def getAdjacencyList(self):
        return self.adj_list

    def getGraph(self):
        return self.graph


    @abstractmethod
    def generateWalk(self):
        pass

    @abstractmethod
    def learnEmbedding(self):
        pass

    @abstractmethod
    def learnNodeEmbedding(self):
        pass


    @abstractmethod
    def getNodeEmbedding(self):
        pass

    @abstractmethod
    def learnEdgeEmbedding(self):
        pass


    @abstractmethod
    def getEdgeEmbedding(self):
        pass